/*
This file is part of CASToR.

    CASToR is free software: you can redistribute it and/or modify it under the
    terms of the GNU General Public License as published by the Free Software
    Foundation, either version 3 of the License, or (at your option) any later
    version.

    CASToR is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along with
    CASToR (in file GNU_GPL.TXT). If not, see <http://www.gnu.org/licenses/>.

Copyright 2017 all CASToR contributors listed below:

    --> current contributors: Thibaut MERLIN, Simon STUTE, Didier BENOIT, Marina FILIPOVIC
    --> past contributors: Valentin VIELZEUF

This is CASToR version 1.1.
*/

/*!
  \file
  \ingroup  image
  \brief    Declaration of class iDeformationElastic
*/

#ifndef IDEFORMATIONELASTIC_HH
#define IDEFORMATIONELASTIC_HH 1

// (Optional) may be required if one want access to the image matrices
#include "oImageSpace.hh"
// Required to automatically add the class in the CASToR code
#include "sAddonManager.hh"
// Mother class of deformation model
#include "vDeformation.hh"

#ifdef CASTOR_ELASTIX
  #include "itkImage.h"
  #include "itkPoint.h"
//  #include "itkRawImageIO.h"
  #include "itkImageFileReader.h"
  #include "itkImageFileWriter.h"
  
  #include "elastixlib.h"
  #include "transformixlib.h"
  #include "elxParameterObject.h"
  #include "itkParameterFileParser.h"
#endif


#ifdef CASTOR_ELASTIX
  using namespace transformix;
  typedef TRANSFORMIX::ParameterMapType RegistrationParametersType;
  typedef itk::ParameterFileParser ParserType;
  typedef itk::ParameterFileParser::ParameterMapType                ParameterMapType;

  enum { Dimension =  3 };
  typedef  float                                      PixelType;
  typedef  itk::Image< PixelType, Dimension >         ImageType;
  typedef  ImageType::IndexType      PixelIndex;
  typedef  ImageType::RegionType     RegionType;
  typedef  ImageType::SizeType       RegionSize;
  typedef  ImageType::IndexType      RegionStart;
  typedef  ImageType::SpacingType    SpacingType;
  typedef  itk::ImageFileReader< ImageType >  ReaderType;
  typedef  itk::ImageFileWriter< ImageType >  WriterType;
#endif


/*!
  \class   iDeformationElastic
  \brief   This class is a squeleton implementation of the pure virtual vDeformation class.
  \details Use this class to implement your own custom deformation model.
*/
class iDeformationElastic : public vDeformation
{
  // -----------------------------------------------------------------------------------------
  // Constructor & Destructor
  public:
    /*!
      \fn      iDeformationElastic::iDeformationElastic
      \brief   Constructor of iDeformationRigid. Simply set all data members to default values.
    */
    iDeformationElastic();
    /*!
      \fn      iDeformationElastic::~iDeformationElastic
      \brief   Destructor of iDeformationRigid. Free memory from all allocated tabs.
    */
    ~iDeformationElastic();


  // -----------------------------------------------------------------------------------------
  // Public member functions
  public:
    // Function for automatic insertion (put the class name as the parameters and do not add semi-colon at the end of the line)
    FUNCTION_DEFORMATION(iDeformationElastic)
    /*!
      \fn      public int iDeformationElastic::ReadAndCheckConfigurationFile()
      \param   const string& a_configurationFile
      \brief   This function is an implementation of the pure virtual mother function. It is used to read options from a configuration file.
      \return  0 if success, other value otherwise.
    */
    int ReadAndCheckConfigurationFile(const string& a_fileOptions);
    /*!
      \fn      public int iDeformationElastic::ReadAndCheckOptionsList()
      \param   const string& a_optionsList
      \brief   This function is an implementation of the pure virtual mother function. It is used to read options from a list of options.
      \return  0 if success, other value otherwise.
    */
    int ReadAndCheckOptionsList(const string& a_listOptions);
    /*!
      \fn      iDeformationElastic::CheckSpecificParameters
      \brief   This function is an implementation of the pure virtual mother function. It is used to
               check parameters of the child deformation model before initialization.
      \return  0 if success, other value otherwise.
    */
    int CheckSpecificParameters();
    /*!
      \fn      public int iDeformationElastic::Initialize()
      \brief   This function is an implementation of the pure virtual mother function. It is used to
               initialize specific stuff to the child deformation model.
      \return  0 if success, other value otherwise.
    */
    int Initialize();
    /*!
      \fn      public void iDeformationElastic::ShowHelp()
      \brief   This function is an implementation of the pure virtual mother function. It is used to
               print out specific help about the deformation model and its options.
    */
    void ShowHelp();
    /*!
      \fn      public virtual int iDeformationElastic::ApplyDeformations()
      \param   FLTNB* ap_inputImage : input image to deform
      \param   FLTNB* ap_outputImage : image in which the output of the deformation should be recovered
      \param   int a_direction : a direction for the deformation to perform (forward or backward)
      \param   int a_defIdx : index of the deformation
      \brief   This function is an implementation of the pure virtual mother function. The actual deformation should be implemented here 
      \return  0 if success, other value otherwise.
    */
    int ApplyDeformations(FLTNB* ap_inputImage, FLTNB* ap_outputImage, int a_direction, int a_defIdx);


  // -----------------------------------------------------------------------------------------
  // Private member functions
  private:


  // -----------------------------------------------------------------------------------------
  // Data members
  private:
    vector<string> m_pathToFwdDeformationFiles; /*!< Containers of the location of forward deformation file. It should be equal to the number of transformations 'm_nbTransformations' */
    vector<string> m_pathToBwdDeformationFiles; /*!< Containers of the location of backward deformation file. It should be equal to the number of transformations 'm_nbTransformations' */
    //ParserType::Pointer file_parser;
    #ifdef CASTOR_ELASTIX
    ImageType::Pointer          ITK_inputImage;
    ImageType::Pointer          ITK_outputImage;
    PixelIndex                  m_PixelIndex;
    RegionType                  m_RegionInputImage;
    RegionSize                  m_RegionSizeInputImage;
    RegionStart                 m_RegionStartInputImage;
    SpacingType                 m_RegionSpacingInputImage;
    #endif
};

// Class for automatic insertion (set here the visible image deformation's name, put the class name as the parameters and do not add semi-colon at the end of the line)
CLASS_DEFORMATION(deformationElastic,iDeformationElastic)

#endif
