/*
This file is part of CASToR.

    CASToR is free software: you can redistribute it and/or modify it under the
    terms of the GNU General Public License as published by the Free Software
    Foundation, either version 3 of the License, or (at your option) any later
    version.

    CASToR is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along with
    CASToR (in file GNU_GPL.TXT). If not, see <http://www.gnu.org/licenses/>.

Copyright 2017 all CASToR contributors listed below:

    --> current contributors: Thibaut MERLIN, Simon STUTE, Didier BENOIT, Marina FILIPOVIC
    --> past contributors: Valentin VIELZEUF

This is CASToR version 1.1.
*/

/*!
  \file
  \ingroup  image
  \brief    Declaration of class iDeformationRigid
*/


#ifndef IDEFORMATIONRIGID_HH
#define IDEFORMATIONRIGID_HH 1

#include "oImageSpace.hh"
#include "sAddonManager.hh"
#include "vDeformation.hh"

/*!
  \class    iDeformationRigid
  \brief    This class performs rigid transformation based on a trilinear interpolation
            It requires ASCII parameter files containing x,y,z transformation vectors
  \details  TODO
*/
class iDeformationRigid : public vDeformation
{
  // -----------------------------------------------------------------------------------------
  // Constructor & Destructor
  public:
    /*!
      \fn      iDeformationRigid::iDeformationRigid
      \brief   Constructor of iDeformationRigid. Simply set all data members to default values.
    */
    iDeformationRigid();
    /*!
      \fn      iDeformationRigid::~iDeformationRigid
      \brief   Destructor of iDeformationRigid. Free memory from all allocated tabs.
    */
    ~iDeformationRigid();


  // -----------------------------------------------------------------------------------------
  // Public member functions
  public:
    // Function for automatic insertion (put the class name as the parameters and do not add semi-colon at the end of the line)
    FUNCTION_DEFORMATION(iDeformationRigid)
    /*!
      \fn      iDeformationRigid::ReadAndCheckConfigurationFile
      \param   a_configurationFile
      \brief   This function is used to read options from a configuration file.
      \return  0 if success, other value otherwise.
    */
    int ReadAndCheckConfigurationFile(const string& a_fileOptions);
    /*!
      \fn      iDeformationRigid::ReadAndCheckOptionsList
      \param   a_optionsList
      \brief   This function is used to read options from a list of options.
               Throw error by defaut for this method, as a file has to be used for initialization
      \return  0 if success, other value otherwise.
    */
    int ReadAndCheckOptionsList(const string& a_listOptions);
    /*!
      \fn      iDeformationRigid::CheckSpecificParameters
      \brief   This function is used to check parameters after the latter
               have been all set using Set functions.
      \return  0 if success, positive value otherwise.
    */
    int CheckSpecificParameters();
    /*!
      \fn      iDeformationRigid::Initialize
      \brief   This function is used to initialize specific stuff in the deformation model.
      \details Allocate memory for image matrices, and initialize images containing transformation parameters
      \return  0 if success, other value otherwise.
    */
    int Initialize();
    /*!
      \fn      iDeformationRigid::ShowHelp
      \brief   This function is used to print out specific help about the deformation and its options.
    */
    void ShowHelp();
    /*!
      \fn      iDeformationRigid::ApplyDeformations
      \param   ap_inputImage : input image to deform
      \param   ap_outputImage : image in which the output of the deformation should be recovered
      \param   a_direction : a direction for the deformation to perform (forward or backward)
      \param   a_defIdx : index of the deformation
      \brief   This function prepares the deformation to perform 
      \details 1. Selects the right deformation parameters file according to the direction and deformation index argument \n
               2. Copy the image to deform in the buffer image of this class \n
               3. Call the deformation function (WarpImage) \n
               4. Copy back the output of the deformation image to the output image matrice passed in argument
      \return  0 if success, other value otherwise.
    */
    int ApplyDeformations(FLTNB* ap_inputImage, FLTNB* ap_outputImage, int a_direction, int a_defIdx);


  // -----------------------------------------------------------------------------------------
  // Private member functions
  private:
    /*!
      \fn      iDeformationRigid::WarpImage
      \param   a_vectorFileIdx : deformation vector file
      \param   ap_inputImage : image on which the deformation should be performed
      \param   ap_outputImage : matrice which recovers the image after deformation
      \brief   This function performs a rigid transformation using x,y,z displacement vectors and trilinear interpolation \n
      \details 1. Load the X,Y,Z transformation vectors from the file passed in argument
               2. Loop on the voxels and compute the new voxels value using trilinear interpolation
      \todo    perhaps add options to load all the vector parameters in RAM during initialization, instead on reading files during reconstruction
      \todo    perhaps use padded image in order to avoid if statements
      \return  0 if success, other value otherwise.
    */
    int WarpImage(string a_vectorFileIdx, double *ap_inputImage, double *ap_outputImage);
    /*!
      \fn      iDeformationRigid::LoadVectorFiles
      \param   a_vectorFileIdx : deformation parameters file
      \brief   Load the transformation parameters from an ASCII file provided in argument
      \return  0 if success, other value otherwise.
    */
    int LoadVectorFiles(string a_vectorFileIdx);


  // -----------------------------------------------------------------------------------------
  // Data members
  private:
    vector<string> 
      m_pathToFwdDeformationFiles; /*!< Containers of the location of forward deformation file.
                                        Nb of elements should be equal to the number of transformations 'm_nbTransformations' */
    vector<string> 
      m_pathToBwdDeformationFiles; /*!< Containers of the location of backward deformation file.
                                        Nb of elements should be equal to the number of transformations 'm_nbTransformations' */
    double* m_OriginalArray;       /*!< Image matrice to get the input image for deformation */
    double* m_OutputArray;         /*!< Image matrice to recover the output image after deformation */
    FLTNB* mp_vectorX;             /*!< Image containing transformation parameters on the X-axis */
    FLTNB* mp_vectorY;             /*!< Image containing transformation parameters on the Y-axis */
    FLTNB* mp_vectorZ;             /*!< Image containing transformation parameters on the Z-axis */
};

// Class for automatic insertion (set here the visible image deformation's name, put the class name as the parameters and do not add semi-colon at the end of the line)
CLASS_DEFORMATION(deformationRigid,iDeformationRigid)

#endif
